///File : DrawableObject.cpp
//------------------------------------------------------------------------------
/*
    This file is part of DevisMaker.

    DevisMaker is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DevisMaker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DevisMaker.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "DrawableObject.hpp"
//------------------------------------------------------------------------------

namespace mdo
{
//------------------------------------------------------------------------------
DrawableObject::DrawableObject(ConfigManager *cfg, ConfigManager *devisCfg, DrawableObject *parrentSizer, const std::string &name)
{
	m_cfg = cfg;
	m_devisCfg = devisCfg;
	m_parentSizer = parrentSizer;
	SetConfigName(name);
	m_isReadyToDraw = false;
}
//------------------------------------------------------------------------------
void DrawableObject::ThrowIfNotOk() const
{
    if( !IsOk() )
    {
        throw ticpp::Exception("DrawableObject is not ok !");
    }
}
//------------------------------------------------------------------------------
void DrawableObject::ThrowIfNotReadyToDraw() const
{
    if( !IsReadyToDraw() )
    {
        throw ticpp::Exception("DrawableObject is not ready to draw !");
    }
}
//------------------------------------------------------------------------------
bool DrawableObject::IsReadyToDraw() const
{
	return (m_isReadyToDraw && IsOk());
}
//------------------------------------------------------------------------------
bool DrawableObject::IsOk() const
{
	return (m_cfgName != "" && m_cfgPrefix != "" && m_typeName != "");
}
//------------------------------------------------------------------------------
double DrawableObject::GetWidthProportion() const
{
    if(m_isWidthGiven)
    {
        return m_widthProportion;
    }
    else
    {
        return -1;
    }
}
//------------------------------------------------------------------------------
/*void DrawableObject::SetWidthProportion(const double &value)
{
    m_widthProportion = value;
}*/
//------------------------------------------------------------------------------
double DrawableObject::GetHeightProportion() const
{
    if(m_isHeightGiven)
    {
        return m_heightProportion;
    }
    else
    {
        return -1;
    }
}
//------------------------------------------------------------------------------
/*void DrawableObject::SetHeightProportion(const double &value)
{
    m_heightProportion = value;
}*/
//------------------------------------------------------------------------------
wxSize DrawableObject::GetSizeReal() const
{
	return m_sizeReal;
}
//------------------------------------------------------------------------------
std::string DrawableObject::GetConfigName() const
{
	return m_cfgName;
}
//------------------------------------------------------------------------------
void DrawableObject::SetConfigName(const std::string &name)
{
    if(name != "")
    {
        m_cfgName = name;
    }
    m_cfgPrefix = m_typeName + std::string("_") + m_cfgName + std::string("_");
}
//------------------------------------------------------------------------------
void DrawableObject::LoadFromCfg(const std::string &name)
{
    SetConfigName(name);
    ThrowIfNotOk();

    m_widthProportion = m_cfg->GetDouble(m_cfgPrefix + "widthProportion");
    m_heightProportion = m_cfg->GetDouble(m_cfgPrefix + "heightProportion");
    m_isWidthGiven = m_cfg->GetBool(m_cfgPrefix + "isWidthGiven");
    m_isHeightGiven = m_cfg->GetBool(m_cfgPrefix + "isHeightGiven");

    m_humanName = m_cfg->GetString(m_cfgPrefix + "humanName");
}
//------------------------------------------------------------------------------
void DrawableObject::SaveToCfg(const std::string &name)
{
    SetConfigName(name);
    ThrowIfNotOk();

    m_cfg->SetDouble(m_cfgPrefix + "widthProportion", m_widthProportion);
    m_cfg->SetDouble(m_cfgPrefix + "heightProportion", m_heightProportion);
    m_cfg->SetBool(m_cfgPrefix + "isWidthGiven", m_isWidthGiven);
    m_cfg->SetBool(m_cfgPrefix + "isHeightGiven", m_isHeightGiven);

    m_cfg->SetString(m_cfgPrefix + "humanName", m_humanName);
}
//------------------------------------------------------------------------------
void DrawableObject::CreateToCfg(const std::string &name)
{
    SetConfigName(name);
    ThrowIfNotOk();

    m_cfg->SetDouble(m_cfgPrefix + "widthProportion", 1);
    m_cfg->SetDouble(m_cfgPrefix + "heightProportion", 1);
    m_cfg->SetBool(m_cfgPrefix + "isWidthGiven", false);
    m_cfg->SetBool(m_cfgPrefix + "isHeightGiven", false);

    m_cfg->SetString(m_cfgPrefix + "humanName", StrWxToStd(_("New Object")));
}
//------------------------------------------------------------------------------
void DrawableObject::RemoveFromCfg(const std::string &name, const bool &withChilds)
{
    SetConfigName(name);
    ThrowIfNotOk();

    m_cfg->RemoveDouble(m_cfgPrefix + "widthProportion");
    m_cfg->RemoveDouble(m_cfgPrefix + "heightProportion");
    m_cfg->RemoveBool(m_cfgPrefix + "isWidthGiven");
    m_cfg->RemoveBool(m_cfgPrefix + "isHeightGiven");

    m_cfg->RemoveString(m_cfgPrefix + "humanName");
}
//------------------------------------------------------------------------------
std::string DrawableObject::GetType() const
{
    return m_typeName;
}
//------------------------------------------------------------------------------
// Special treeCtrl
wxTreeItemId DrawableObject::GetTreeItemId() const
{
    return m_treeItemId;
}
//------------------------------------------------------------------------------
void DrawableObject::SetTreeItemId(const wxTreeItemId &treeItemId)
{
    m_treeItemId = treeItemId;
}
//------------------------------------------------------------------------------
DrawableObject* DrawableObject::GetParent()
{
    return m_parentSizer;
}
//------------------------------------------------------------------------------
void DrawableObject::SetParent(DrawableObject* parentSizer)
{
    m_parentSizer = parentSizer;
}
//------------------------------------------------------------------------------
bool DrawableObject::IsChildOfSizer(const DrawableObject *sizer) const
{
    if( m_parentSizer == sizer ) // C'est mon pére
    {
        return true;
    }
    else if( m_parentSizer == NULL ) // On est arrivé au root ==> Pas mon pére
    {
        return false;
    }

    // Si non : on demande à mes parents
    return m_parentSizer->IsChildOfSizer(sizer);
}
//------------------------------------------------------------------------------
void DrawableObject::EditPropertyDialog(DialogProperty *window)
{
    wxPanel *PagePanel;
    wxBoxSizer *PageSizer;

    // General
    PagePanel = new wxPanel(window->GetNotebook(), wxID_ANY);
    PageSizer = new wxBoxSizer(wxVERTICAL);

    window->AddParamCtrl(type_check_enable_next, PagePanel, PageSizer, m_cfgPrefix + "isWidthGiven", _("Use a defined width proportion"));
    window->AddParamCtrl(type_double_per_cent, PagePanel, PageSizer, m_cfgPrefix + "widthProportion", _("Parent width proportion used by the object (%) :"));

    window->AddParamCtrl(type_check_enable_next, PagePanel, PageSizer, m_cfgPrefix + "isHeightGiven", _("Use a defined height proportion"));
    window->AddParamCtrl(type_double_per_cent, PagePanel, PageSizer, m_cfgPrefix + "heightProportion", _("Parent height proportion used by the object (%) :"));

    PagePanel->SetSizer(PageSizer);
    window->GetNotebook()->AddPage(PagePanel, _("Proportion"));
}
// HumanName
std::string DrawableObject::GetHumanName() const
{
    return m_humanName;
}
//------------------------------------------------------------------------------
void DrawableObject::SetHumanName(const std::string &name)
{
    m_humanName = name;
    SaveToCfg();
}
//------------------------------------------------------------------------------
void DrawableObject::AutoChangeConfigName()
{
    // Nom à changer...
    const std::string prefix = GetType();
    int suffix = 1;
    while(DrawableObjectExistsInCfg(m_cfg, GetType(), prefix + nbToString(suffix)))
    {
        suffix++;
    }

    SetConfigName(prefix + nbToString(suffix));
}
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//---------------------------- Fonctions ---------------------------------------
//------------------------------------------------------------------------------
wxString ArrangeTextForWidth(const wxDC *dc, wxString str, const int &width)
{
    // V2 : On Ajoute des mots jusqu'à ce que ce soit ok pour la ligne ==> Plus rapide pour les longs textes (> 2 lignes environ).
    wxString finalText;
    wxString tempLine;
    int nextWidth = 0;

    while(!str.IsEmpty()) // Tant qu'il reste des mots à caser
    {
        while(!str.IsEmpty())
        {
            int nextSpacePos = str.Find(' ');
            if(nextSpacePos == wxNOT_FOUND) // Si plus d'espace, on va jusqu'à la fin.
            {
                nextSpacePos = str.Len() - 1;
            }

            wxString wordToMesure = str.Mid(0, nextSpacePos);
            if(wordToMesure[wordToMesure.Len() - 1] == ' ') // Si le mot se fini par un espace, on le supprime pour la mesure. (toujours le cas sauf si le mot va jusqu'à la fin du texte).
            {
                wordToMesure = wordToMesure.Mid(0, wordToMesure.Len() - 1);
            }
            nextWidth = GetTextSize(dc, tempLine + wordToMesure).GetWidth();

            if(nextWidth <= width)
            {
                // On ajoute
                tempLine += str.Mid(0, nextSpacePos + 1);

                // On supprime de str
                str = str.Mid(nextSpacePos + 1);
            }
            else if(tempLine.IsEmpty()) // Si la ligne est vide, on ajoute des lettres
            {
                do // Ajout au moins une fois
                {
                    // On ajoute
                    tempLine += str[0];

                    // On supprime de str
                    str = str.Mid(1);
                }
                while( !str.IsEmpty() && GetTextSize(dc, tempLine + str.Mid(0, 1)).GetWidth() < width);

                break; // On enregistre la ligne.
            }
            else
            {
                break; // On enregistre la ligne.
            }
        }

        // Ajout de la ligne...
        if(!str.IsEmpty()) // Si on est pas à la derniére ligne...
        {
            if(tempLine[tempLine.Len() - 1] == ' ') // Si le mot se fini par un espace, on le supprime pour la mesure. (toujours le cas sauf si le mot va jusqu'à la fin du texte).
            {
                tempLine = tempLine.Mid(0, tempLine.Len() - 1);
            }

            tempLine << _T("\n");
        }
        finalText << tempLine;
        tempLine = _T("");
    }

    // ok
    return finalText;

    // V1 : on enléve des mots jusqu'à ce que ce soit ok pour la ligne.

    /*wxString finalText;
    wxString tempWords;
    while(str != _T("")) // Tant qu'il reste des mots à caser
    {
        int currentWidth = GetTextSize(dc, str).GetWidth();
        while(currentWidth > width && str.Len() > 1)
        {
            // Tant que la chaine str est trop longue, on enléve des mots...
            int wordPos = str.Find(' ', true);
            if(wordPos != wxNOT_FOUND && wordPos != 0) // Si il y a au moins un mot... et que l'on ne coup epas l'esace du debut.
            {
                tempWords = str.Mid(wordPos) + tempWords;
                str = str.Mid(0, wordPos);
            }
            else // Si pas de mots ==> Lettres par lettres
            {
                tempWords = str.Right(1) + tempWords;
                str = str.Mid(0, str.Len() - 1);
            }

            currentWidth = GetTextSize(dc, str).GetWidth();
        }

        // str est à la bonne taille ==> On la stock
        finalText << str << wxChar('\n');

        // On remet les mots sauvegardé dans str pour une nouvelle ligne.
        str = tempWords;
        tempWords = _T("");

    }

    // On enléve le '\n' à la fin de finalText...
    finalText = finalText.Mid(0, finalText.Len() - 1);

    // ok
    return finalText;*/
}
//------------------------------------------------------------------------------
wxSize GetTextSize(const wxDC *dc, wxString str)
{
    wxSize txtSize(0, 0);

    while(str != _T(""))
    {
        wxString toWrite;
        if(str.Find('\n') != wxNOT_FOUND)
        {
            toWrite = str(0, str.Find('\n'));
            if(toWrite == _T(""))
            {
                 toWrite +=   + _T(" ");// Ajout de " " pour ne pas avoir de ligne vide en cas de double saute ("\n\n"), qui ont une hateur de 0.
            }
            str = str(str.Find('\n') + 1, str.Len());
        }
        else
        {
            toWrite = str;
            str = _T("");
        }

        txtSize.y += dc->GetTextExtent(toWrite).GetHeight();

        if(dc->GetTextExtent(toWrite).GetWidth() > txtSize.x)
        {
            txtSize.x = dc->GetTextExtent(toWrite).GetWidth();
        }
    }

    return txtSize;
}
//------------------------------------------------------------------------------
void DrawText(wxDC *dc, wxPoint pos, wxString str, const int &vAlign, const int &hAlign )
{
    if(str == _T(""))
    {
        return;
    }

    wxSize textSize;
    if(vAlign != vAlign_Left || hAlign != hAlign_Top) // Si besoin de calculer la taille...
    {
        textSize = GetTextSize(dc, str);
    }

    if(vAlign == vAlign_Center)
    {
        pos.x -= textSize.x / 2;
    }
    else if(vAlign == vAlign_Right)
    {
        pos.x -= textSize.x;
    }

    if(hAlign == hAlign_Center)
    {
        pos.y -= textSize.y / 2;
    }
    else if(hAlign == hAlign_Bottom)
    {
        pos.y -= textSize.y;
    }

    while(str != _T(""))
    {
        wxString toWrite;
        if(str.Find('\n') != wxNOT_FOUND)
        {
            toWrite = str(0, str.Find('\n'));
            if(toWrite == _T(""))
            {
                 toWrite +=   + _T(" ");// Ajout de " " pour ne pas avoir de ligne vide en cas de double saute ("\n\n"), qui ont une hateur de 0.
            }
            str = str(str.Find('\n') + 1, str.Len());
        }
        else
        {
            toWrite = str;
            str = _T("");
        }

        dc->DrawText(toWrite, pos.x, pos.y);

        pos.y += dc->GetTextExtent(toWrite).GetHeight();
    }
}
//------------------------------------------------------------------------------
bool DrawableObjectExistsInCfg(const ConfigManager *cfg, const std::string &type, const std::string &name)
{
    return (cfg->ExistsDouble(type + std::string("_") + name + std::string("_") + std::string("widthProportion")));
}
//------------------------------------------------------------------------------
void DCSelectColorFont(wxDC *dc, const StructWxColorFont &colorFont)
{
    dc->SetFont( colorFont.wxfont );
    dc->SetTextForeground( colorFont.wxcolor );
}
//------------------------------------------------------------------------------
} // namespace mdo
